from zerotest.request import Request


req = Request(scheme="http", method="GET", params={"hello": "test"}, host="example.com", path="/test",
              headers={"Auth": "FOR_TEST"}, data="ok")


def test_url():
    assert req.url == 'http://example.com/test'


def test_endpoint():
    assert req.endpoint == 'http://example.com'


def test_endpoint_setter():
    endpoint = req.endpoint
    req.endpoint = "https://another-example.com"
    assert req.endpoint == "https://another-example.com"
    assert req.host == "another-example.com"
    assert req.scheme == "https"
    req.endpoint = endpoint


def test__str__():
    assert str(req) == """[GET]http://example.com/test
{'Auth': 'FOR_TEST'}
ok"""



# coding=utf8
from __future__ import unicode_literals

import six


def ensure_unicode(content):
    if content is None:
        return content
    elif six.PY2:
        return '{}'.format(content)
    elif isinstance(content, bytes):
        # python3
        return content.decode('utf-8')

    return content



# encoding: utf-8
import re

INVALID_METHOD_R = re.compile(r'[^a-zA-Z0-9_]')
MULTI_UNDERSCORE_R = re.compile(r'[_]+')


def get_name_from_request(request):
    return '{}_{}'.format(
        request.method.lower(),
        _path_to_func_name(request.path)
    )


def _path_to_func_name(path):
    """
    Generate a valid python function name for a given request path.

    The valid characters for a Python 2 function name are defined as follows:

    identifier ::=  (letter|"_") (letter | digit | "_")*
    letter     ::=  lowercase | uppercase
    lowercase  ::=  "a"..."z"
    uppercase  ::=  "A"..."Z"
    digit      ::=  "0"..."9"

    :param path: The HTTP request path string.
    :type path: unicode
    :rtype: unicode
    """
    if path == '/':
        return 'root'

    # Replace any occurances of more than one underscore with a single
    # instance. (ex: path = '/#anchor' -> '__anchor')
    return MULTI_UNDERSCORE_R.sub(
        '_',
        # Replace any invalid method name characters with an underscore.
        INVALID_METHOD_R.sub(
            '_',
            path
        )
    ).strip('_')


def dict_to_param_style_code(d):
    return ', '.join('{0}={1}'.format(k, repr(v)) for k, v in d.items())



import logging
import threading

from zerotest.record.formatter import Formatter
from zerotest.utils.queue_helper import Queue, Empty

LOG = logging.getLogger(__name__)


class HTTPRecorder(object):
    """
    format and record tunnel request/response to file
    """

    def __init__(self, filepath):
        """
        :param filepath: new record will append to file if filepath exists
        :return:
        """
        self.filepath = filepath
        self._running = False
        self._closing = False
        self._service_thread = None
        self._queue = Queue()
        self._formatter = Formatter()

    def start_service(self):
        """
        start recorder service
        :return:
        """
        self._service_thread = threading.Thread(target=self._loop_work)
        self._running = True
        LOG.info("recorder service start, store http record to '{}'".format(self.filepath))
        self._service_thread.start()

    def _loop_work(self):
        import codecs
        with codecs.open(self.filepath, 'a+', 'utf-8') as record_file:
            while True:
                task = self._queue.get()
                LOG.debug("receive task %s", task)
                if task is None:
                    record_file.close()
                    return

                self._formatter.write_record(record_file, task[0], task[1])
                record_file.flush()
                LOG.debug("writen task %s", task)

    def record_http(self, request, response):
        """
        async method, put request, response into task queue
        :return:
        """
        if not self._closing:
            LOG.debug("record http %s, %s", request, response)
            self._queue.put((request, response))

    def close(self):
        """
        graceful close, wait util queue empty
        :return:
        """
        if self._running:
            LOG.debug("closing...")
            self._closing = True
            self._queue.put(None)
            LOG.debug("wait task complete...")
            self._service_thread.join()
        else:
            raise RuntimeError("current service is not running")

        LOG.info("closed, record file: '{}'".format(self.filepath))

    def shutdown(self):
        """
        shutdown service, threw out items already in queue
        :return:
        """
        LOG.debug("shutdown...")
        self._closing = True
        try:
            while True:
                self._queue.get_nowait()
        except Empty:
            pass

        self.close()



# encoding: utf-8
from collections import OrderedDict
from zerotest.request import Request
from zerotest.utils.generator_helper import (
    get_name_from_request,
    dict_to_param_style_code
)


#: Common HTTP request path characters that are invalid when used for
#: a python function name.
INVALID_CHARS = (
    '*',
    '-',
    '#',
    '/',
    '.'
)


def test_get_name_from_request():
    req = Request(path='/', method='GET')
    assert get_name_from_request(req) == 'get_root'
    req = Request(path='/the/world', method='PUT')
    assert get_name_from_request(req) == 'put_the_world'

    for invalid_char in INVALID_CHARS:
        assert(
            get_name_from_request(
                Request(
                    path='/bad_{0}_trailing'.format(invalid_char),
                    method='GET'
                )
            ) == 'get_bad_trailing'
        )


def test_dict_to_param_style_code():
    test_dict = OrderedDict()
    test_dict['param_a'] = 'A'
    test_dict['test_c'] = 'C'
    assert dict_to_param_style_code(test_dict) == "param_a='A', test_c='C'"



import logging

from zerotest.common import DEBUG

if not DEBUG:
    logging.getLogger("werkzeug").setLevel(logging.ERROR)
    logging.getLogger("zerotest.record.http_recorder").setLevel(logging.ERROR)



from __future__ import unicode_literals

import logging

import werkzeug.wrappers

from zerotest.request import Request
from zerotest.response import Response
from zerotest.utils.http_helper import response_with_response

LOG = logging.getLogger(__name__)


class Forwarder(object):
    def __init__(self, forward_url):
        self._forward_url = forward_url
        self._on_forward_complete_callbacks = []

    def __call__(self, environ, start_response):
        # pop incorrect content length, I don't know why
        if not environ.get('CONTENT_LENGTH'):
            environ.pop('CONTENT_LENGTH', None)

        request = werkzeug.wrappers.Request(environ)

        headers = {k: v for k, v in request.headers if k not in ('Host',)}
        LOG.debug("forward to [%s]%s, headers: -----%s-----, data %s",
                  request.method, self._forward_url, headers, request.data)
        forward_request = Request(method=request.method, headers=headers, data=request.data,
                                  params=request.query_string, path=request.path, endpoint=self._forward_url)
        response = forward_request.send_request()
        forward_response = Response.from_requests_response(response)
        self.trigger_on_forward_complete(forward_request, forward_response)
        return response_with_response(response, start_response)

    def on_forward_complete(self, callback):
        self._on_forward_complete_callbacks.append(callback)

    def trigger_on_forward_complete(self, request, response):
        for callback in self._on_forward_complete_callbacks:
            callback(request, response)



from zerotest.utils.io_helper import StringIO

from zerotest.record.formatter import Formatter
from zerotest.request import Request
from zerotest.response import Response

req = Request(scheme="http", method="get", params="query_string=here", host="example.com", path="/test",
              headers={"just": "header"}, data="request")
res = Response(status=200, headers={"responsed": "header"}, body="response")
formatter = Formatter()


def test_formatter():
    writable = StringIO()
    formatter.write_record(writable, req, res)
    readable = StringIO(writable.getvalue())
    request, response = formatter.read_record(readable)
    assert req.__dict__ == request.__dict__
    assert res.__dict__ == response.__dict__



# coding=utf8
from __future__ import unicode_literals

"""
Mocked API server
"""
import threading
import time
import json

import requests
from werkzeug.serving import run_simple
from werkzeug.wrappers import Request, Response

from tests.mock import pickup_port
from zerotest.utils.generator_helper import get_name_from_request
from zerotest.utils.url_helper import urljoin


class Server(object):
    def __init__(self):
        self.count = 0
        self.host = None
        self.port = None
        self.thread = None
        self.running = False

    def dispatch_request(self, request):
        handler = getattr(self, get_name_from_request(request), None)
        if handler:
            return handler(request)
        else:
            return Response("non route", status=404)

    def get_count(self, request):
        return Response(json.dumps(dict(count=self.count)),
                        content_type='application/json')

    def post_echo(self, request):
        return Response(request.data,
                        content_type=request.headers['content_type'])

    def post_raw_to_json(self, request):
        data = request.data
        return Response(data,
                        content_type='application/json')

    def get_chinese_hello_world(self, request):
        return Response(json.dumps(dict(count=self.count, content="你好,世界")),
                        content_type='application/json')

    def delete_shutdown_server(self, request):
        func = request.environ.get('werkzeug.server.shutdown')
        if func is None:
            raise RuntimeError('Not running with the Werkzeug Server')
        func()
        return Response("shutdown")

    def __call__(self, environ, start_response):
        request = Request(environ)
        self.count += 1
        environ['HTTP_COUNT'] = self.count
        return self.dispatch_request(request)(environ, start_response)

    def start(self, host, port):
        self.host = host
        self.port = port
        self.running = True
        run_simple(host, port, self)

    @property
    def url(self):
        return 'http://{}:{}'.format(self.host, self.port)

    def start_daemon(self, host, port):
        self.thread = threading.Thread(target=lambda: self.start(host, port))
        self.thread.start()

    def shutdown(self):
        if self.running:
            requests.delete(urljoin(self.url, 'shutdown_server'))
            self.running = False
            self.thread.join(30)

    def start_mock_server(self):
        port = pickup_port()
        self.start_daemon('127.0.0.1', port)
        test_count = 10
        while test_count > 0:
            try:
                if self.running:
                    r = requests.get(urljoin(self.url, '/count'))
                    if r.status_code == 200:
                        return

                    print("wait server start, response", r.status_code, r.text)
            except requests.exceptions.ConnectionError:
                pass

            time.sleep(1)
            test_count -= 1
        else:
            self.shutdown()
            raise RuntimeError("start mock server timeout")



from zerotest.utils import http_helper


def test_dict_to_wsgi_headers():
    data = {"blue": "yellow", "yellow": "red", "red": "blue"}
    assert sorted(http_helper.dict_to_wsgi_headers(data)) == sorted(
        [("blue", "yellow"), ("yellow", "red"), ("red", "blue")])



import logging


MatchError = AssertionError

LOG = logging.getLogger(__name__)

_SERIALIZABLE_CONTENT_TYPE = {'application/json': 'json'}


class ResponseMatcher(object):
    def __init__(self, ignore_headers=None, ignore_all_headers=False, ignore_fields=None):
        """
        :param ignore_headers: ignored headers when match response
        :param ignore_fields: ignored body fields, only work when response content-type is serializable type
        :return:
        """
        ignore_headers = ignore_headers or []
        self._ignore_headers = set(map(lambda h: h.upper(), ignore_headers))
        self._ignore_all_headers = ignore_all_headers
        self._ignore_fields = ignore_fields

    def _compare_status(self, r1, r2):
        assert r1.status == r2.status

    def __remove_ignore_headers(self, headers):
        return {k.upper(): headers[k] for k in headers if
                k.upper() not in self._ignore_headers}

    def _compare_headers(self, expect, real):
        if not self._ignore_all_headers:
            expect_headers = self.__remove_ignore_headers(expect.headers)
            real_headers = self.__remove_ignore_headers(real.headers)
            assert expect_headers == real_headers

    def __delete_ignore_fields(self, content):
        from zerotest.utils.data_helper import delete_path_from_dict
        for field_path in self._ignore_fields:
            if field_path in content:
                content.pop(field_path, None)
            else:
                delete_path_from_dict(content, field_path)

    def _handle_content_type_json(self, content):
        import json
        try:
            content = json.loads(content)
        except:
            LOG.error("detected json response, but raise a error in decoding")
            raise

        self.__delete_ignore_fields(content)
        return content

    def _compare_body(self, r1, r2):
        r1_content_type = r1.get_header('Content-Type')
        r2_content_type = r2.get_header('Content-Type')
        assert r1_content_type == r2_content_type
        r1_content = r1.body
        r2_content = r2.body
        content_type = _SERIALIZABLE_CONTENT_TYPE.get(r1_content_type)
        if self._ignore_fields and content_type:
            handler = getattr(self, '_handle_content_type_{}'.format(content_type))
            if handler:
                r1_content = handler(r1_content)
                r2_content = handler(r2_content)

        assert r1_content == r2_content

    def match_responses(self, expect, real):
        """
        compare requests
        :type expect: zerotest.response.Response
        :type real: zerotest.response.Response
        :return:
        """
        for attr in ('status', 'headers', 'body'):
            compare_func = '_compare_{}'.format(attr)
            getattr(self, compare_func)(expect, real)



from __future__ import unicode_literals

from zerotest.utils.generator_helper import dict_to_param_style_code


class Response(object):
    def __init__(self, status=None, headers=None, body=None):
        self.status = status
        self.headers = headers
        self.body = body

    @staticmethod
    def from_requests_response(response):
        res = Response(status=response.status_code, body=response.text,
                       headers=dict(response.headers))
        # make dict key/value to unicode
        import json
        res.__dict__ = json.loads(json.dumps(res.__dict__))
        return res

    def get_header(self, header):
        for k, v in self.headers.items():
            if k.upper() == header.upper():
                return v

    def __eq__(self, other):
        if type(other) != Response:
            return False

        return self.__dict__ == other.__dict__

    def __str__(self):
        return """{status}
{headers}
{body}""".format(**self.__dict__)

    def __repr__(self):
        return '{}({})'.format(Response.__name__, dict_to_param_style_code(self.__dict__))



from copy import copy


def dict_to_wsgi_headers(headers):
    wsgi_headers = []
    for k, v in headers.items():
        wsgi_headers.append((k, v))

    return wsgi_headers


def response_with_response(response, start_response):
    headers = copy(response.headers)
    # remove transfer encoding
    for k in ('Content-Encoding', 'Transfer-Encoding'):
        headers.pop(k, None)
    start_response("{} {}".format(response.status_code, response.reason), dict_to_wsgi_headers(headers))
    return response



from zerotest.response import Response


def test_response__str__():
    response = Response(200, {"just test": "hope pass"}, "happy test!")
    assert str(response) == """200
{'just test': 'hope pass'}
happy test!"""



from __future__ import unicode_literals

"""
test renderer of python
"""

from collections import defaultdict
from jinja2 import Template
from zerotest.utils.generator_helper import get_name_from_request, dict_to_param_style_code


class Renderer(object):
    def __init__(self, options, match_options):
        self.options = options
        self.match_options = match_options

    def prepare(self, records):
        cases = []
        func_names = defaultdict(int)
        endpoint = self.options.get('endpoint', None)
        ignore_all_headers = self.match_options.get('ignore_all_headers', None)
        ignore_headers = self.match_options.get('ignore_headers', None)
        if not ignore_all_headers and ignore_headers:
            ignore_headers = set(map(lambda h: h.upper(), ignore_headers))

        for (req, res) in records:
            func_name = get_name_from_request(req)
            count = func_names[func_name]
            func_names[func_name] = count + 1
            if count > 0:
                func_name = '{}_{}'.format(func_name, count)

            if ignore_all_headers:
                content_type = res.get_header('content-type')
                if content_type:
                    res.headers = {'content-type': content_type}
                else:
                    res.headers = {}
            elif ignore_headers:
                res.headers = {k: v for k, v in res.headers.items() if
                               k.upper() not in ignore_headers}

            case_info = dict(request=req, response=res, func_name=func_name)
            cases.append(case_info)
        return cases

    def render(self, records):
        cases = self.prepare(records)
        t = Template(_TEMPLATE)
        result = t.render(options=self.options, match_params=dict_to_param_style_code(self.match_options), cases=cases)
        return result


_TEMPLATE = """# Generated by zerotest
from __future__ import unicode_literals
from zerotest.request import Request
from zerotest.response import Response
from zerotest.response_matcher import ResponseMatcher


matcher = ResponseMatcher({{ match_params }})
verify_ssl = {{ options.verify_ssl or False }}
{% if options.endpoint %}
endpoint = {{ options.endpoint.__repr__() }}
{% endif %}
{% for c in cases %}
def test_{{ c.func_name }}():
    request = {{ c.request.__repr__() }}
    {% if options.endpoint %}
    request.endpoint = endpoint
    {% endif %}
    real = Response.from_requests_response(request.send_request(verify=verify_ssl))
    expect = {{ c.response.__repr__() }}
    matcher.match_responses(expect, real)

{% endfor %}
"""



from zerotest.response import Response
from zerotest.response_matcher import ResponseMatcher, MatchError


def raised_error(func, err_type):
    raised = False
    try:
        func()
    except err_type:
        raised = True

    return raised


def test_compare():
    matcher = ResponseMatcher()
    r1 = Response(status=200, headers={"a": "1"}, body="ok")
    r2 = Response(status=200, headers={"a": "1"}, body="ok")
    f = lambda: matcher.match_responses(r1, r2)
    assert not raised_error(f, MatchError)
    r2.body = "not ok"
    assert raised_error(f, MatchError)
    r2.body = "ok"
    r2.status = 201
    assert raised_error(f, MatchError)
    r2.status = 200
    r2.headers['b'] = 2
    assert raised_error(f, MatchError)
    del r2.headers['b']
    assert not raised_error(f, MatchError)


def test_ignore_headers_compare():
    matcher = ResponseMatcher(ignore_headers=['c', 'd', 'e'])
    r1 = Response(status=200, headers={"a": "1", "c": "not"}, body="ok")
    r2 = Response(status=200, headers={"a": "1", "c": "same"}, body="ok")
    f = lambda: matcher.match_responses(r1, r2)
    assert not raised_error(f, MatchError)
    r1.headers['d'] = "ignored should"
    assert not raised_error(f, MatchError)
    r2.headers['e'] = "ignored should"
    assert not raised_error(f, MatchError)
    r1.headers['g'] = "not ignored"
    assert raised_error(f, MatchError)


def test_ignore_all_headers_compare():
    matcher = ResponseMatcher(ignore_all_headers=True)
    r1 = Response(status=200, headers={"a": "1", "c": "not"}, body="ok")
    r2 = Response(status=200, headers={"a": "1", "c": "same", "d": 3}, body="ok")
    f = lambda: matcher.match_responses(r1, r2)
    assert not raised_error(f, MatchError)


def test_ignore_fields_compare():
    import json
    matcher = ResponseMatcher(ignore_fields=['some_record.created_at', 'id'])
    f = lambda: matcher.match_responses(r1, r2)
    r1 = Response(status=200, headers={}, body=json.dumps({"id": 1, "name": "test"}))
    r2 = Response(status=200, headers={}, body=json.dumps({"id": 2, "name": "test"}))
    # not work if not content-type
    assert raised_error(f, MatchError)
    r1.headers['content-type'] = 'application/json'
    r2.headers['CONTENT-TYPE'] = 'application/json'
    assert not raised_error(f, MatchError)

    r1.body = json.dumps({"some_record.created_at": 'now', 'record': {'created_at': '111'}})
    r2.body = json.dumps({"record": {'created_at': '111'}})
    assert not raised_error(f, MatchError)
    r1.body = json.dumps({"some_record.created_at": 'now', 'some_record': {'created_at': '111'}})
    r2.body = json.dumps({"some_record": {'created_at': '111'}})
    assert raised_error(f, MatchError)

    r1.body = json.dumps({"some_record.created_at": 'now'})
    r2.body = json.dumps({"some_record": {'created_at': '112'}})
    assert raised_error(f, MatchError)

    # test key path under list
    r1.body = json.dumps({"some_record": [{'created_at': '1123', 'id': '0'}]})
    r2.body = json.dumps({"some_record": [{'created_at': '112', 'id': '1'}]})
    assert raised_error(f, MatchError)
    r1.body = json.dumps({"some_record": [{'created_at': '1123', 'id': '0'}, {'created_at': '3322'}]})
    r2.body = json.dumps({"some_record": [{'created_at': '112', 'id': '1'}, {'created_at': 'tomorrow'}]})
    assert raised_error(f, MatchError)

    r1.body = json.dumps({"some_record": [{'created_at': '1123', 'id': '1'}, {'created_at': '3322'}]})
    r2.body = json.dumps({"some_record": [{'created_at': '112', 'id': '1'}, {'created_at': 'tomorrow'}]})
    assert not raised_error(f, MatchError)



import logging

from zerotest.record.formatter import Formatter


LOG = logging.getLogger(__name__)


class Generator(object):
    """
    Generator test scripts from record data
    """

    def __init__(self, filepath, options=None, match_options=None):
        self._filepath = filepath
        default_options = dict(verify_ssl=False)
        if options:
            default_options.update(options)

        self._options = options or dict()
        self._formatter = Formatter()
        self._match_options = match_options or dict()

    def generate(self):
        """
        return None if records data not found
        :return:
        """

        import codecs

        records = []
        with codecs.open(self._filepath, 'r', 'utf-8') as record_file:
            while True:
                result = self._formatter.read_record(record_file)
                if not result:
                    break
                records.append(result)

        from zerotest.generator.python.render import Renderer

        if len(records) == 0:
            return None

        renderer = Renderer(options=self._options,
                            match_options=self._match_options)

        return renderer.render(records)



import subprocess
import shlex

from tests import DEBUG


def call_process(cmd):
    try:
        from subprocess import DEVNULL
    except ImportError:
        import os
        DEVNULL = open(os.devnull, 'wb')
    if DEBUG:
        out = None
        close_fds = False
    else:
        out = DEVNULL
        close_fds = True

    args = shlex.split(cmd)
    return subprocess.call(args, stdout=out, stderr=out, close_fds=close_fds)



from __future__ import unicode_literals
import argparse
import logging
import os
import sys
import tempfile

from zerotest.common import init_logging_config
from zerotest.utils.url_helper import urlparse

DESCRIPTION = """
Capture HTTP request/response and convert to test code.
"""
init_logging_config()

LOG = logging.getLogger(__name__)


class CLI(object):
    def __init__(self):
        self._parser = None
        self._parse_result = None

    def _init_arg_parser(self):
        parser = argparse.ArgumentParser(description=DESCRIPTION)
        subparsers = parser.add_subparsers(help='sub-command help', dest='subparser_name')

        server_parser = subparsers.add_parser('server', help='start zerotest local proxy server')
        server_parser.add_argument('url', help="target url, example: http://example.com")
        server_parser.add_argument('-f', '--file', help="file path to save record data, default: [random path]")
        server_parser.add_argument('-b', '--bind', help="local bind address, default: 127.0.0.1")
        server_parser.add_argument('-p', '--port', help="local port, default: 7000")

        generate_parser = subparsers.add_parser('generate', help='generate test code from record data')
        self._add_match_options_to_parser(generate_parser)

        replay_parser = subparsers.add_parser('replay', help='replay test from record data')
        self._add_match_options_to_parser(replay_parser)
        replay_parser.add_argument('-t', '--pytest', help="pass options to pytest, example: -t='-vv'", nargs='*')

        self._parser = parser

    @staticmethod
    def _add_match_options_to_parser(parser):
        parser.add_argument('file', help="path of record data file")
        parser.add_argument('--endpoint', help="replace requests endpoint, https://example.com")
        parser.add_argument('--ignore-headers', help="list of headers ignore in response matching",
                            nargs='*')
        parser.add_argument('--ignore-all-headers', help="skip headers match", dest="ignore_all_headers",
                            action='store_true'),
        parser.add_argument('--ignore-fields',
                            help="list of fields ignore in response matching,"
                                 " only work on serializable content-type,"
                                 " example: --ignore-fields a b c.d(explained as d under c)",
                            nargs='*')
        parser.add_argument('--verify-ssl', help="enable ssl verify", dest="verify_ssl", action="store_true")
        parser.add_argument('--no-verify-ssl', help="disable ssl verify", dest="verify_ssl",
                            action="store_false")
        parser.set_defaults(verify_ssl=False)
        parser.set_defaults(ignore_all_headers=False)

    def run(self, argv=sys.argv[1:]):
        """
        run cli
        :param argv:
        :return: cli exit code
        """
        self._init_arg_parser()
        self._parse_result = self._parser.parse_args(argv)
        if not self.verify_parse_result():
            return 1
        return getattr(self, 'command_{}'.format(self._parse_result.subparser_name))()

    def verify_parse_result(self):
        cli_options = self.get_cli_options('endpoint', 'url')
        endpoint = cli_options.get('endpoint')
        if endpoint:
            parsed_url = urlparse(endpoint)
            if not parsed_url.hostname or not parsed_url.scheme:
                LOG.error("invalid endpoint '{}'".format(endpoint))
                return False

        url = cli_options.get('url')
        if url:
            parsed_url = urlparse(url)
            if not parsed_url.hostname:
                LOG.error("invalid url '{}'".format(url))
                return False

        return True

    def get_cli_options(self, *keys):
        options = {k: getattr(self._parse_result, k) for k in keys if
                   getattr(self._parse_result, k, None)}
        return options

    def command_server(self):
        """
        sub-command start
        :return:
        """
        from zerotest.app import App

        forward_url = self._parse_result.url
        filepath = self._parse_result.file

        if not filepath:
            _, filepath = tempfile.mkstemp()
        else:
            if os.path.exists(filepath):
                LOG.warning("file '{}' is exists, new record will append to the file".format(filepath))

        app = App(forward_url, filepath)
        host = self._parse_result.bind or '127.0.0.1'
        port = int(self._parse_result.port or 7000)
        app.run(host, port)

    def command_replay(self):
        """
        sub-command replay
        run record file
        :return:
        """
        import tempfile
        import pytest

        generator = self._generator_from_command()
        pytest_args = self._parse_result.pytest or []
        with tempfile.NamedTemporaryFile('w+', suffix='.py') as f:
            f.write(generator.generate())
            f.flush()
            return pytest.main([f.name] + pytest_args)

    def _generator_from_command(self):
        from zerotest.generator.generator import Generator
        filepath = self._parse_result.file
        options = self.get_cli_options('endpoint', 'verify_ssl')
        match_options = self.get_cli_options('ignore_headers', 'ignore_fields', 'ignore_all_headers')
        generator = Generator(filepath, options=options, match_options=match_options)
        return generator

    def command_generate(self):
        """
        sub-command generate
        :return:
        """
        generator = self._generator_from_command()
        print(generator.generate())


def main():
    code = CLI().run() or 0
    exit(code)


if __name__ == '__main__':
    main()



from zerotest.generator.python.render import Renderer
from zerotest.request import Request
from zerotest.response import Response


def test_renderer():
    options = dict()
    match_options = dict()
    records = list()
    records.append((Request(scheme='http', method='GET', host='for_test.org', path='/',
                            headers=dict(header_1='1'), data='just_for_test'),
                    Response(status=200, headers=dict(content_type="text"), body='ok')))
    records.append((Request(scheme='https', method='POST', host='for_test.org', path='/second_request',
                            headers=dict(header_a='a'), data='second_request'),
                    Response(status=200, headers=dict(content_type="text"), body='ok')))
    renderer = Renderer(options=options, match_options=match_options)

    cases = [
        dict(request=Request(scheme='http', method='GET', host='for_test.org', path='/',
                             headers=dict(header_1='1'), data='just_for_test')
             , response=Response(status=200, headers=dict(content_type="text"), body='ok')
             , func_name='get_root'),
        dict(request=Request(scheme='https', method='POST', host='for_test.org', path='/second_request',
                             headers=dict(header_a='a'), data='second_request')
             , response=Response(status=200, headers=dict(content_type="text"), body='ok')
             , func_name='post_second_request'),
    ]

    assert renderer.prepare(records) == cases


def test_renderer_with_options():
    options = dict(endpoint='http://test.com')
    match_options = dict()
    records = list()
    records.append((Request(scheme='http', method='GET', host='for_test.org', path='/',
                            headers=dict(header_1='1'), data='just_for_test'),
                    Response(status=200, headers=dict(content_type="text"), body='ok')))
    records.append((Request(scheme='https', method='POST', host='for_test.org', path='/second_request',
                            headers=dict(header_a='a'), data='second_request'),
                    Response(status=200, headers=dict(content_type="text"), body='ok')))
    renderer = Renderer(options=options, match_options=match_options)

    cases = [
        dict(request=Request(scheme='http', method='GET', host='for_test.org', path='/',
                             headers=dict(header_1='1'), data='just_for_test')
             , response=Response(status=200, headers=dict(content_type="text"), body='ok')
             , func_name='get_root'),
        dict(request=Request(scheme='https', method='POST', host='for_test.org', path='/second_request',
                             headers=dict(header_a='a'), data='second_request')
             , response=Response(status=200, headers=dict(content_type="text"), body='ok')
             , func_name='post_second_request'),
    ]

    assert renderer.prepare(records) == cases


def test_renderer_with_ignore_headers():
    options = dict()
    match_options = dict(ignore_headers=('date', 'auth'))
    records = list()
    records.append((Request(scheme='http', method='GET', host='for_test.org', path='/',
                            headers=dict(header_1='1'), data='just_for_test'),
                    Response(status=200, headers=dict(content_type="text", date='yesterday', auth='secret'),
                             body='ok')))
    records.append((Request(scheme='https', method='POST', host='for_test.org', path='/second_request',
                            headers=dict(header_a='a'), data='second_request'),
                    Response(status=200, headers=dict(content_type="text", date='today'), body='ok')))
    renderer = Renderer(options=options, match_options=match_options)

    cases = [
        dict(request=Request(scheme='http', method='GET', host='for_test.org', path='/',
                             headers=dict(header_1='1'), data='just_for_test')
             , response=Response(status=200, headers=dict(content_type="text"), body='ok')
             , func_name='get_root'),
        dict(request=Request(scheme='https', method='POST', host='for_test.org', path='/second_request',
                             headers=dict(header_a='a'), data='second_request')
             , response=Response(status=200, headers=dict(content_type="text"), body='ok')
             , func_name='post_second_request'),
    ]

    assert renderer.prepare(records) == cases


def test_renderer_with_ignore_all_headers():
    options = dict()
    match_options = dict(ignore_all_headers=True)
    records = list()
    records.append((Request(scheme='http', method='GET', host='for_test.org', path='/',
                            headers=dict(header_1='1'), data='just_for_test'),
                    Response(status=200, headers={"content-type": "text", "date": 'yesterday', "auth": 'secret'},
                             body='ok')))
    records.append((Request(scheme='https', method='POST', host='for_test.org', path='/second_request',
                            headers=dict(header_a='a'), data='second_request'),
                    Response(status=200, headers={"content-type": "text", "date": 'today'}, body='ok')))
    renderer = Renderer(options=options, match_options=match_options)

    cases = [
        dict(request=Request(scheme='http', method='GET', host='for_test.org', path='/',
                             headers=dict(header_1='1'), data='just_for_test')
             , response=Response(status=200, headers={"content-type": "text"}, body='ok')
             , func_name='get_root'),
        dict(request=Request(scheme='https', method='POST', host='for_test.org', path='/second_request',
                             headers=dict(header_a='a'), data='second_request')
             , response=Response(status=200, headers={"content-type": "text"}, body='ok')
             , func_name='post_second_request'),
    ]

    assert renderer.prepare(records) == cases

    options = dict()
    match_options = dict(ignore_all_headers=True)
    records = list()
    records.append((Request(scheme='http', method='GET', host='for_test.org', path='/',
                            headers=dict(header_1='1'), data='just_for_test'),
                    Response(status=200, headers={"date": 'yesterday', "auth": 'secret'},
                             body='ok')))
    records.append((Request(scheme='https', method='POST', host='for_test.org', path='/second_request',
                            headers=dict(header_a='a'), data='second_request'),
                    Response(status=200, headers={"date": 'today'}, body='ok')))
    renderer = Renderer(options=options, match_options=match_options)

    cases = [
        dict(request=Request(scheme='http', method='GET', host='for_test.org', path='/',
                             headers=dict(header_1='1'), data='just_for_test')
             , response=Response(status=200, headers={}, body='ok')
             , func_name='get_root'),
        dict(request=Request(scheme='https', method='POST', host='for_test.org', path='/second_request',
                             headers=dict(header_a='a'), data='second_request')
             , response=Response(status=200, headers={}, body='ok')
             , func_name='post_second_request'),
    ]

    assert renderer.prepare(records) == cases




import os
import logging

DEBUG = os.getenv('ZEROTEST_DEBUG', 'false').lower() == 'true'


def init_logging_config():
    if DEBUG:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)
        # disable requests log
        logging.getLogger("requests").setLevel(logging.WARNING)



from zerotest.forwarder import Forwarder


request = None
response = None


def test_callback():
    def callback(req, res):
        global request
        global response
        request = req
        response = res

    forwarder = Forwarder('http://example.com')
    forwarder.on_forward_complete(callback)
    fake_request = object()
    fake_response = object()
    forwarder.trigger_on_forward_complete(fake_request, fake_response)
    assert request == fake_request
    assert response == fake_response



from __future__ import unicode_literals

import requests

from zerotest.utils.generator_helper import dict_to_param_style_code
from zerotest.utils.url_helper import urlparse
from zerotest.utils.encode_helper import ensure_unicode


class Request(object):
    def __init__(self, scheme=None, method=None, params=None, host=None, path=None, headers=None, data=None,
                 endpoint=None):
        self.scheme = scheme
        self.method = method
        self.headers = headers
        self.host = host
        self.path = path
        self.params = ensure_unicode(params)
        self.data = ensure_unicode(data)
        if endpoint:
            self.endpoint = endpoint

    @property
    def endpoint(self):
        return "{scheme}://{host}".format(**self.__dict__)

    @endpoint.setter
    def endpoint(self, endpoint):
        parsed = urlparse(endpoint)
        self.scheme = parsed.scheme
        if parsed.port:
            host = "{}:{}".format(parsed.hostname, parsed.port)
        else:
            host = parsed.hostname
        self.host = host

    @property
    def url(self):
        return "{}{}".format(self.endpoint, self.path)

    def send_request(self, verify=True):
        return requests.request(self.method, self.url, headers=self.headers,
                                params=self.params, data=self.data,
                                stream=True, allow_redirects=False, verify=verify)

    def __eq__(self, other):
        if type(other) != Request:
            return False

        return self.__dict__ == other.__dict__

    def __str__(self):
        return """[{method}]{url}
{headers}
{data}""".format(method=self.method, url=self.url, headers=self.headers, data=self.data)

    def __repr__(self):
        return '{}({})'.format(Request.__name__,
                               dict_to_param_style_code({k: v for k, v in self.__dict__.items() if v}))



from zerotest.cli import CLI


class AttributeDict(object):
    def __init__(self, **kwargs):
        self.__dict__ = kwargs


def test_verify_parse_result():
    cli = CLI()
    cli._parse_result = AttributeDict(url='invalid.url')
    assert not cli.verify_parse_result()
    cli._parse_result = AttributeDict(url='http://valid.url')
    assert cli.verify_parse_result()
    cli._parse_result = AttributeDict(endpoint='invalid.url')
    assert not cli.verify_parse_result()
    cli._parse_result = AttributeDict(endpoint='http://valid.url')
    assert cli.verify_parse_result()



from __future__ import unicode_literals

import json

from zerotest.request import Request
from zerotest.response import Response


class Formatter(object):
    def write_record(self, writeable, request, response):
        record = dict(request=request.__dict__, response=response.__dict__)
        writeable.write(json.dumps(record))
        writeable.write("\n")

    def read_record(self, readable):
        line = readable.readline()
        if not line:
            return None
        record = json.loads(line)
        request = Request()
        request.__dict__.update(record['request'])
        response = Response()
        response.__dict__.update(record['response'])
        return request, response



def delete_path_from_dict(d, path):
    """
    delete key path from dict
    :param d: dict
    :param path: key path example: a.b.c
    :return:
    """
    if not d:
        return
    paths = path.split('.')
    parent = d
    last_path = ''
    for i, p in enumerate(paths):
        if not d:
            return
        elif isinstance(d, list):
            for item in d:
                delete_path_from_dict(item, '.'.join(paths[i:]))
            return

        last_path = p
        parent = d
        d = d.get(p)

    parent.pop(last_path, None)



#!/usr/bin/env python
from __future__ import unicode_literals

import logging
import subprocess
import sys

from setuptools import Command
from setuptools import find_packages, setup

from zerotest.common import init_logging_config
from zerotest.utils.encode_helper import ensure_unicode

init_logging_config()

LOG = logging.getLogger(__name__)

try:
    long_description = ensure_unicode(subprocess.check_output(["pandoc", "README.md", "-f", "markdown", "-t", "rst"]))
except (OSError, subprocess.CalledProcessError) as e:
    LOG.error("call pandoc error: %s", e)
    LOG.warning("failed convert README from markdown to rst, read as text")
    with open('README.md') as fd:
        long_description = fd.read()


class TestCommand(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(['tests'])
        sys.exit(errno)


setup(name='zerotest',
      version='1.1.1',
      long_description=long_description,
      description="Lazy guy's testing tool, test your API server like a boss.",
      author='Hari Jiang',
      author_email='hari.jiang@outlook.com',
      url='https://github.com/jjyr/zerotest',
      license='MIT',
      cmdclass={'test': TestCommand},
      platforms=['unix', 'linux', 'osx'],
      packages=find_packages(),
      install_requires=[
          'requests>=2.2.1',
          'Werkzeug>=0.10.4',
          'jinja2>=2.8',
          'pytest>=2.8.3',
          'six>=1.10.0',
      ],
      entry_points={
          'console_scripts': [
              'zerotest=zerotest.cli:main',
          ],
      },
      classifiers=[
          'License :: OSI Approved :: MIT License',
          'Operating System :: POSIX',
          'Operating System :: MacOS :: MacOS X',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Topic :: Software Development :: Libraries',
          'Topic :: Software Development :: Testing',
          'Topic :: Utilities',
      ],
      )



# coding=utf8
from __future__ import unicode_literals

import json
import os
import unittest

import requests

from tests.helper.process import call_process
from tests.mock.mock_server import Server
from tests.mock.zerotest_proxy import Proxy
from zerotest.utils.url_helper import urljoin

server = None
proxy = None
zerotest_cmd = "python zerotest/cli.py"


def init_test_env():
    global server, proxy
    try:
        server = Server()
        server.start_mock_server()
        proxy = Proxy()
        proxy.start_server(server.url)
    except:
        destroy_test_env()
        raise


def destroy_test_env():
    global server, proxy
    if proxy:
        proxy.shutdown()
    if server:
        server.shutdown()


class TestZerotest(unittest.TestCase):
    def setUp(self):
        init_test_env()

    def tearDown(self):
        destroy_test_env()
        if proxy.data_file:
            os.remove(proxy.data_file)

    def test_proxy(self):
        # test 404
        res = requests.get(urljoin(proxy.url, 'echo'))
        assert res.status_code == 404

        res = requests.post(urljoin(server.url, 'echo'), data='who am I?')
        res.raise_for_status()
        assert res.text == 'who am I?'

        res = requests.post(urljoin(proxy.url, 'echo'), data='who am I?')
        res.raise_for_status()
        assert res.text == 'who am I?'

        res = requests.get(urljoin(proxy.url, 'count'))
        res.raise_for_status()
        assert isinstance(res.json()['count'], int)

        data = dict(name="test json convert", magic_number=7)
        res = requests.post(urljoin(proxy.url, 'raw/to.json'), data=json.dumps(data))
        res.raise_for_status()
        assert data == res.json()

        res = requests.get(urljoin(proxy.url, 'chinese/hello-world'))
        res.raise_for_status()
        assert res.json()['content'] == '你好,世界'
        proxy.shutdown()

    def test_replay(self):
        # test replay
        replay = lambda args: call_process(zerotest_cmd + args + " -t='-vv'")
        assert replay(" replay not_exist_file") != 0
        assert replay(" replay {}".format(proxy.data_file)) != 0
        assert replay(" replay {} --ignore-all-headers".format(proxy.data_file)) != 0
        assert replay(" replay {} --ignore-all-headers --ignore-fields count".format(proxy.data_file)) == 0
        assert replay(" replay {} --ignore-headers date --ignore-fields count".format(proxy.data_file)) == 0



"""
proxy of zerotest proxy server
"""

import signal
import tempfile
import time
from subprocess import Popen

import requests

from tests.mock import pickup_port
from zerotest.utils.url_helper import urljoin


class Proxy(object):
    def __init__(self):
        self.port = None
        self._process = None
        _, self.data_file = tempfile.mkstemp()

    @property
    def url(self):
        return 'http://127.0.0.1:{}'.format(self.port)

    @property
    def running(self):
        if not self._process:
            return False
        self._process.poll()
        return self._process.returncode is None

    def shutdown(self):
        if self.running:
            self._process.send_signal(signal.SIGINT)
            self._process.wait()

    def start_server(self, url):
        """
        :param url: tested server
        :return:
        """
        port = pickup_port()
        self._process = Popen(['python', 'zerotest/cli.py', 'server', '-p', str(port), '-f', self.data_file, url])
        self.port = port
        test_count = 10
        while test_count > 0:
            try:
                if self.running:
                    r = requests.get(urljoin(self.url, '/count'))
                    if r.status_code == 200:
                        return
            except requests.exceptions.ConnectionError:
                pass

            print("wait proxy start,", test_count)
            time.sleep(1)
            test_count -= 1
        else:
            self.shutdown()
            raise RuntimeError("start zerotest proxy timeout")



import os
import tempfile

from zerotest.record.formatter import Formatter
from zerotest.record.http_recorder import HTTPRecorder
from zerotest.request import Request
from zerotest.response import Response


req = Request(scheme="http", method="get", params="query_string=here", host="example.com", path="/test",
              headers={"just": "header"}, data="request")
res = Response(status=200, headers={"responsed": "header"}, body="response")


def test_http_recorder():
    _, filepath = tempfile.mkstemp()
    formatter = Formatter()
    recorder = HTTPRecorder(filepath)
    try:
        recorder.start_service()
        recorder.record_http(req, res)
    finally:
        recorder.close()

    with open(filepath, 'r') as readable:
        request, response = formatter.read_record(readable)
        assert req.__dict__ == request.__dict__
        assert res.__dict__ == response.__dict__

    os.remove(filepath)
